/*
 * Copyright (c) 2007, CustomWare Asia Pacific
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of "CustomWare Asia Pacific" nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.randombits.confluence.metadata.impl;

import com.atlassian.confluence.core.ContentEntityObject;
import com.atlassian.confluence.labels.Label;
import com.atlassian.confluence.pages.Attachment;
import com.atlassian.confluence.search.lucene.extractor.HTMLSearchableTextExtractor;
import com.atlassian.user.User;
import org.randombits.confluence.metadata.MetadataManager;
import org.randombits.confluence.metadata.MetadataStorage;
import org.xml.sax.SAXException;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import com.atlassian.confluence.plugins.index.api.Extractor2;
import com.atlassian.confluence.plugins.index.api.FieldDescriptor;

/**
 * Extracts information for data fields into the search engine.
 *
 * @author David Peterson
 * @author Abi Dzar
 */
public class ScaffoldDataIndexer implements Extractor2 {

    // private static final Logger LOG = Logger.getLogger( ScaffoldDataIndexer.class );

    private static final DateFormat DATE_FORMAT = new SimpleDateFormat( "dd MM MMM MMMM yyyy hh:mm a HH:mm" );
    private static final DateFormat LUCENE_DATE_FORMAT = new SimpleDateFormat("yyyyMMddHHmmssSSS");

    private final MetadataManager metadataManager;

    public ScaffoldDataIndexer(MetadataManager metadataManager) {
        this.metadataManager = metadataManager;
    }
    @Override
    public Collection<FieldDescriptor> extractFields(Object searchable)
    {
        StringBuffer defaultSearchableText = new StringBuffer();
        Collection<FieldDescriptor> document = new ArrayList<FieldDescriptor>();
        if ( searchable instanceof ContentEntityObject ) {
            // LOG.debug("Searchable is a ContentEntityObject.");
            ContentEntityObject content = (ContentEntityObject) searchable;
            MetadataStorage metadata = metadataManager.loadWritableData( content );
            if ( metadata != null ) {
                // LOG.debug("Searchable properties were found");
                processBox( metadata, document, defaultSearchableText );
            }
        }
        return document;
    }

    @Override
    public StringBuilder extractText(Object searchable)
    {
        return new StringBuilder(0);
    }

    private void processBox( MetadataStorage metadata, Collection<FieldDescriptor> document, StringBuffer defaultSearchableText ) {
        for ( String name : metadata.nameSet() ) {
            Object value = metadata.getObject( name, null );

            processValue( metadata, document, defaultSearchableText, name, value );
        }
    }

    private void processValue( MetadataStorage metadata, Collection<FieldDescriptor> document, StringBuffer defaultSearchableText,
                               String name, Object value ) {
        try {
            if ( value != null ) {
                if ( value instanceof String ) {
                    try {
                        value = HTMLSearchableTextExtractor.stripTags( (String) value );
                    } catch ( SAXException e ) {
                        // Do nothing - index as is.
                    }
                    document.add( new FieldDescriptor( name, (String) value, FieldDescriptor.Store.NO, FieldDescriptor.Index.ANALYZED ) );
                } else if ( value instanceof Date ) {
                    value = handleDateValue( name, (Date) value, document );
                } else {
                    if ( value instanceof Number || value instanceof Boolean ) {
                        value = value.toString();
                    } else if ( value instanceof ContentEntityObject ) {
                        value = ( (ContentEntityObject) value ).getTitle();
                    } else if ( value instanceof Attachment ) {
                        value = ( (Attachment) value ).getFileName();
                    } else if ( value instanceof User ) {
                        value = ( (User) value ).getName();
                    } else if ( value instanceof Label ) {
                        value = ( (Label) value ).getName();
                    }

                    if ( value instanceof String ) {
                        document.add( new FieldDescriptor( name, (String) value, FieldDescriptor.Store.YES, FieldDescriptor.Index.ANALYZED ) );
                    }
                }

                if ( value instanceof List<?> ) {
                    for ( Object i : (List<?>) value ) {
                        processValue( metadata, document, defaultSearchableText, name, i );
                    }
                } else if ( value instanceof Map<?, ?> ) {
                    metadata.openBox( name );
                    processBox( metadata, document, defaultSearchableText );
                    metadata.closeBox();
                } else {
                    defaultSearchableText.append( value ).append( " " );
                }
            }
        } catch ( Exception e ) {
            // Do nothing.
        }
    }

    /**
     * @param key      The key to name the data FieldDescriptor.
     * @param value    The value of the date FieldDescriptor.
     * @param document The document being generated for indexing.
     * @return The converted value of the date FieldDescriptor.
     */
    private String handleDateValue( String key, Date value, Collection<FieldDescriptor> document ) {
        document.add( new FieldDescriptor( key, LUCENE_DATE_FORMAT.format(value),
                FieldDescriptor.Store.YES, FieldDescriptor.Index.ANALYZED ) );
        return DATE_FORMAT.format( value );
    }
}